;;;; aasummarizer.scm  -- summarize AA               -*- coding: utf-8 -*-
;;
;; $Id: aasummarizer.scm,v 1.110 2009/12/05 08:20:11 hos Exp $
;;

;;
;; DB:
;;
;; "/top" -> list of title IDs
;; title-ID -> human readable title-name, list of subtitle-IDs,
;;             related page (optional)
;; subtitle-ID -> human readable subtitle-name, list of article-IDs,
;;                heading (optional), script (optional)
;; article-ID -> src text, registered date
;;
;; ID of title   : /top/[0-9a-zA-Z-_]+
;; ID of subtitle: title/[0-9a-zA-Z-_]+
;; ID of article : subtitle/[0-9]+
;;


(define-module aasummarizer)
(select-module aasummarizer)

(export aasummarizer-main
        cgi-path
        common-header
        page-editable?
        )

(use srfi-1)
(use srfi-11)
(use srfi-13)
(use srfi-19)
(use util.list)
(use gauche.parameter)
(use gauche.charconv)
(use text.html-lite)
(use www.cgi)
(use dbm)

(autoload dbm.fsdbm <fsdbm>)
(autoload file.util make-directory* copy-file file-equal? file-device=?)
(autoload text.tree write-tree)
(autoload text.tr string-tr)
(autoload rfc.uri uri-encode-string)
(autoload sxml.tools sxml:string->xml)


;; db

(define db (make-parameter #f))

(define (with-db db-type path need-write? proc)
  (parameterize ((db (dbm-open db-type
                               :path path
                               :rw-mode (if need-write? :write :read))))
    (dynamic-wind
        (lambda () #f)
        proc
        (lambda ()
          (unless (dbm-closed? (db))
            (dbm-close (db)))))))

(define (get-db-data key)
  (and (dbm-exists? (db) key)
       (call-with-input-string (dbm-get (db) key) read)))

(define (put-db-data! key val)
  (dbm-put! (db) key (call-with-output-string (pa$ write val))))

(define (put-data! data)
  (put-db-data! (id-of data) (alist-of data)))

(define (remove-db-data! key)
  (dbm-delete! (db) key))

(define (remove-data! data)
  (remove-db-data! (id-of data)))


;; cgi params

(define cgi-params (make-parameter #f))

(define (get-param key def)
  (cgi-get-parameter
   key (cgi-params)
   :default def
   :convert (lambda (val)
              (or (and (string? val)
                       (let ((v (ces-convert val
                                             (cgi-output-character-encoding))))
                         (if (string-incomplete? v)
                             (ces-convert val "*JP")
                             v)))
                  val))))


;; data classes

(define-class <root-data> ()
  ((titles :accessor titles-of :init-keyword :titles)
   ))

(define-class <title> ()
  ((id :accessor id-of :init-keyword :id)
   (name :accessor name-of :init-keyword :name)
   (subtitles :accessor subtitles-of :init-keyword :subtitles)
   (related-page :accessor related-page-of :init-keyword :related-page)
   (keywords :accessor keywords-of :init-keyword :keywords)
   ))

(define-class <subtitle> ()
  ((id :accessor id-of :init-keyword :id)
   (title :accessor title-of :init-keyword :title)
   (name :accessor name-of :init-keyword :name)
   (articles :accessor articles-of :init-keyword :articles)
   (heading :accessor heading-of :init-keyword :heading)
   (script :accessor script-of :init-keyword :script)
   ))

(define-class <article> ()
  ((id :accessor id-of :init-keyword :id)
   (header :accessor header-of :init-keyword :header)
   (footer :accessor footer-of :init-keyword :footer)
   (body :accessor body-of :init-keyword :body)
   (style :accessor style-of :init-keyword :style)
   (c-date :accessor c-date-of :init-keyword :c-date)
   (m-date :accessor m-date-of :init-keyword :m-date)
   ))


;; common

(define (base-id-of data)
  ((#/\/([^\/]*)$/ (id-of data)) 1))

(define (base-path-of data)
  ((#/^(.*)\/[^\/]*$/ (id-of data)) 1))


;; root-data

(define root-id "/top")

(define-method id-of ((root <root-data>))
  root-id)

(define (get-root-data)
  (let ((data (get-db-data root-id)))
    (make <root-data>
      :titles (if data
                  (assq-ref data 'titles)
                  '())
      )))

(define-method alist-of ((root <root-data>))
  `((titles . ,(titles-of root))
    ))


;; title

(define (get-title root title-id)
  (let* ((id #`",(id-of root)/,|title-id|")
         (data (get-db-data id)))
    (and data
         (make <title>
           :id id
           :name (assq-ref data 'name)
           :subtitles (assq-ref data 'subtitles)
           :related-page (assq-ref data 'related-page)
           :keywords (assq-ref data 'keywords)
           ))))

(define-method alist-of ((title <title>))
  `((name . ,(name-of title))
    (subtitles . ,(subtitles-of title))
    ,@(if (related-page-of title)
          `((related-page . ,(related-page-of title)))
          '())
    ,@(if (keywords-of title)
          `((keywords . ,(keywords-of title)))
          '())
    ))

(define (add-title! root title-id title-name)
  (let* ((id #`",(id-of root)/,|title-id|")
         (title (make <title>
                 :id id
                 :name title-name
                 :subtitles '()
                 :related-page #f
                 :keywords #f
                 )))
    (set! (titles-of root)
          (cons title-id (titles-of root)))
    (put-data! title)
    (put-data! root)
    title))

(define (update-title! title title-name)
  (set! (name-of title) title-name)
  (put-data! title))

(define (update-title-id! root title-id new-id)
  (let* ((title (get-title root title-id))
         (new-title (add-title! root new-id (name-of title))))
    (update-title-relpage! new-title (related-page-of title))
    (update-title-keywords! new-title (keywords-of title))
    (move-title! root new-id title-id)
    (copy-subtitles! title new-title)
    )
  (remove-title! root title-id))

(define (update-title-relpage! title relpage)
  (set! (related-page-of title) relpage)
  (put-data! title))

(define (update-title-keywords! title keywords)
  (set! (keywords-of title) keywords)
  (put-data! title))

(define (move-title! root title-id next-title-id)
  (receive (before after)
    (break (pa$ equal? next-title-id) (titles-of root))
    (set! (titles-of root)
          (append (delete title-id before)
                  `(,title-id)
                  (delete title-id after)))
    (put-data! root)))

(define (remove-title! root title-id)
  (let ((title (get-title root title-id)))
    (set! (titles-of root)
          (delete title-id (titles-of root)))
    (put-data! root)
    (for-each (lambda (subtitle-id)
                (remove-subtitle! title subtitle-id))
              (subtitles-of title))
    (remove-data! title)
    ))


;; subtitle

(define (get-subtitle title subtitle-id)
  (let* ((id #`",(id-of title)/,|subtitle-id|")
         (data (get-db-data id)))
    (and data
         (make <subtitle>
           :id id
           :title title
           :name (assq-ref data 'name)
           :articles (assq-ref data 'articles)
           :heading (assq-ref data 'heading "")
           :script (assq-ref data 'script "")
           ))))

(define-method alist-of ((subtitle <subtitle>))
  `((name . ,(name-of subtitle))
    (articles . ,(articles-of subtitle))
    (heading . ,(heading-of subtitle))
    (script . ,(script-of subtitle))
    ))

(define (add-subtitle! title subtitle-id subtitle-name)
  (let* ((id #`",(id-of title)/,|subtitle-id|")
         (subtitle (make <subtitle>
                     :id id
                     :title title
                     :name subtitle-name
                     :articles '()
                     :heading ""
                     :script ""
                     )))
    (set! (subtitles-of title)
          (receive (before after)
            (break (pa$ equal? "other") (subtitles-of title))
            (append before `(,subtitle-id) after)))
    (put-data! subtitle)
    (put-data! title)
    subtitle))

(define (update-subtitle! subtitle subtitle-name)
  (set! (name-of subtitle) subtitle-name)
  (put-data! subtitle))

(define (update-subtitle-id! title subtitle-id new-id)
  (let* ((subtitle (get-subtitle title subtitle-id))
         (new-subtitle (add-subtitle! title new-id (name-of subtitle))))
    (update-subtitle-heading! new-subtitle
                              (heading-of subtitle) (script-of subtitle))
    (move-subtitle! title new-id subtitle-id)
    (copy-articles! subtitle new-subtitle)
    )
  (remove-subtitle! title subtitle-id))

(define (update-subtitle-heading! subtitle heading script)
  (set! (heading-of subtitle) heading)
  (set! (script-of subtitle) script)
  (put-data! subtitle))

(define (move-subtitle! title subtitle-id next-subtitle-id)
  (receive (before after)
    (break (pa$ equal? next-subtitle-id) (subtitles-of title))
    (set! (subtitles-of title)
          (append (delete subtitle-id before)
                  `(,subtitle-id)
                  (delete subtitle-id after)))
    (put-data! title)))

(define (remove-subtitle! title subtitle-id)
  (let ((subtitle (get-subtitle title subtitle-id)))
    (set! (subtitles-of title)
          (delete subtitle-id (subtitles-of title)))
    (put-data! title)
    (for-each (lambda (article-id)
                (remove-article! subtitle article-id))
              (articles-of subtitle))
    (remove-data! subtitle)
    ))

(define (copy-subtitles! src-title dst-title)
  (for-each
   (lambda (subtitle-id)
     (let* ((subtitle (get-subtitle src-title subtitle-id))
            (new-subtitle (add-subtitle! dst-title subtitle-id
                                         (name-of subtitle))))
       (update-subtitle-heading! new-subtitle
                                 (heading-of subtitle) (script-of subtitle))
       (copy-articles! subtitle new-subtitle)
       ))
   (subtitles-of src-title)))


;; article

(define date-format "~Y/~m/~d ~H:~M:~S ~z")

(define (get-article subtitle article-id)
  (let* ((id #`",(id-of subtitle)/,|article-id|")
         (data (get-db-data id)))
    (and data
         (make <article>
           :id id
           :header (assq-ref data 'header)
           :footer (assq-ref data 'footer)
           :body (assq-ref data 'body)
           :style (assq-ref data 'style (aa-default-style))
           :c-date (string->date (assq-ref data 'c-date) date-format)
           :m-date (string->date (assq-ref data 'm-date) date-format)
           ))))

(define-method alist-of ((article <article>))
  `((header . ,(header-of article))
    (footer . ,(footer-of article))
    (body . ,(body-of article))
    (style . ,(style-of article))
    (c-date . ,(date->string (c-date-of article) date-format))
    (m-date . ,(date->string (m-date-of article) date-format))
    ))

(define (add-article! subtitle header footer body style)
  (let* ((article-id
          (if (null? (articles-of subtitle))
              "1"
              (number->string
               (+ 1 (string->number (last (articles-of subtitle)))))))
         (id #`",(id-of subtitle)/,|article-id|")
         (article (make <article>
                    :id id
                    :header header
                    :footer footer
                    :body body
                    :style style
                    :c-date (current-date)
                    :m-date (current-date)
                    )))
    (set! (articles-of subtitle)
          (append (articles-of subtitle) `(,article-id)))
    (put-data! article)
    (put-data! subtitle)
    article))

(define (update-article! article header footer body style)
  (set! (header-of article) header)
  (set! (footer-of article) footer)
  (set! (body-of article) body)
  (set! (style-of article) style)
  (set! (m-date-of article) (current-date))
  (put-data! article))

(define (remove-article! subtitle article-id)
  (let ((article (get-article subtitle article-id)))
    (set! (articles-of subtitle)
          (delete article-id (articles-of subtitle)))
    (put-data! subtitle)
    (remove-data! article)
    ))

(define (copy-articles! src-subtitle dst-subtitle)
  (for-each
   (lambda (article-id)
     (let* ((article (get-article src-subtitle article-id))
            (new-article (add-article! dst-subtitle
                                       (header-of article)
                                       (footer-of article)
                                       (body-of article)
                                       (style-of article))))
       (set! (c-date-of new-article) (c-date-of article))
       (set! (m-date-of new-article) (m-date-of article))
       (put-data! new-article)
       ))
   (articles-of src-subtitle)))

;; writer utils

(define page-editable? (make-parameter #t))

(define pre-tlist-proc (make-parameter #f))
(define post-tlist-proc (make-parameter #f))

(define pre-alist-proc (make-parameter #f))
(define post-alist-proc (make-parameter #f))

(define export-dest (make-parameter #f))
(define est-export-dest (make-parameter #f))
(define pre-export-proc (make-parameter #f))
(define post-export-proc (make-parameter #f))

(define page-title (make-parameter #f))
(define page-path (make-parameter #f))
(define page-base-url (make-parameter #f))
(define page-base-path (make-parameter #f))
(define page-css (make-parameter #f))
(define page-js (make-parameter #f))
(define page-content-type (make-parameter #f))
(define page-tmp-path (make-parameter #f))

(define page-keywords (make-parameter #f))
(define page-description (make-parameter #f))

(define top-page-writer (make-parameter #f))

(define aa-default-style (make-parameter #f))
(define aa-style-alist (make-parameter #f))

(define robots-noindex? (make-parameter #f))

(define (fix-path path)
  (if (#/^\/|^https?:/ path)
      path
      #`",(page-base-path)/,|path|"))

(define (cgi-path . cmds)
  (let ((cmd-pairs
         (let loop ((pairs '())
                    (cmds cmds)
                    (var? #t))
           (if (null? cmds)
               (reverse pairs)
               (loop (if var?
                         (cons (car cmds) pairs)
                         (cons (cons (car pairs) (car cmds)) (cdr pairs)))
                     (cdr cmds)
                     (not var?))))))
    (if (page-editable?)
        (string-append (page-path)
                       "?"
                       (string-join (map (lambda (p) #`",(car p)=,(cdr p)")
                                         cmd-pairs)
                                    "&"))
        (let ((v (map cdr cmd-pairs)))
          (apply string-append
                 (page-base-path)
                 (string-join (cdr v) "/" 'prefix)
                 "/"
                 (cond ((or (string=? (car v) "f")
                            (string=? (car v) "sf"))
                        '())
                       ((string=? (car v) "ra")
                        '("r.xml"))
                       (else
                        (list (car v) ".html"))))))))

(define (get-text-param key desc)
  (let ((val (get-param key #f)))
    (if (not (string? val))
        (error #`",|desc| is not specified"))
    (string-delete val #\cr)))

(define (get-text-param* key)
  (let ((val (get-param key #f)))
    (and (string? val)
         val)))

(define (get-title-id) (get-text-param "t" "Title"))
(define (get-subtitle-id) (get-text-param "s" "Subtitle"))
(define (get-subtitle-id*) (get-text-param* "s"))
(define (get-article-id) (get-text-param "a" "Article"))
(define (get-edit-name) (get-text-param "n" "Name"))
(define (get-edit-id) (get-text-param "ni" "ID"))
(define (get-header-text) (get-text-param "h" "Header"))
(define (get-footer-text) (get-text-param "f" "Footer"))
(define (get-body-text) (get-text-param "b" "Body"))
(define (get-style-text) (get-text-param "st" "Style"))
(define (get-heading-text) (get-text-param "h" "Heading"))
(define (get-script-text) (get-text-param "r" "Script"))
(define (get-relpage-text) (get-text-param "r" "Related page"))
(define (get-keywords-text) (get-text-param "k" "Keywords"))

(define (success-writer path)
  `(,(html-doctype :type :html-4.01-strict)
    ,(html:html
      (html:head
       (html:meta
        :http-equiv "refresh"
        :content #`"0;URL=,|path|"))
      (html:body
       :class "success"
       (html:p "Successfully completed")
       (html:p (html:a :href path "Reload"))
       ))))

;; escape only '<' and '>'
(define (html-escape-lt-gt)
  (port-for-each (lambda (c)
                   (case c
                     ((#\<) (display "&lt;"))
                     ((#\>) (display "&gt;"))
                     (else (display c))
                     ))
                 read-char))

(define (html-escape-lt-gt-string string)
  (with-string-io (x->string string) html-escape-lt-gt))

;; escape only '<', '>', and '&'
(define (html-escape-lt-gt-amp)
  (port-for-each (lambda (c)
                   (case c
                     ((#\<) (display "&lt;"))
                     ((#\>) (display "&gt;"))
                     ((#\&) (display "&amp;"))
                     (else (display c))
                     ))
                 read-char))

(define (html-escape-lt-gt-amp-string string)
  (with-string-io (x->string string) html-escape-lt-gt-amp))

(define (textarea-escape-string string)
  (let ((string (html-escape-lt-gt-amp-string string)))
    (if (#/^[\r\n]/ string)
        #`"\n,|string|"
        string)))

(define html-aa-escape-string html-escape-lt-gt-string)

(define (jscript-escape)
  (port-for-each (lambda (c)
                   (case c
                     ((#\<) (display "\\x3c"))
                     ((#\>) (display "\\x3e"))
                     ((#\&) (display "\\x26"))
                     ((#\") (display "\\\""))
                     ((#\\) (display "\\\\"))
                     ((#\nl) (display "\\r\\n"))
                     (else (display c))
                     ))
                 read-char))

(define (jscript-escape-string string)
  (with-string-io (x->string string) jscript-escape))

(define (string-jpci<? x y)
  (let ((normalize (lambda (s)
                     (string-tr s
                                "ぁ-ん零一二三四五六七八九A-Z"
                                "ァ-ン0-9a-z"))))
    (string<? (normalize x) (normalize y))))

(define (htmlize str)
  (fold (lambda (l res) `(,l ,(if (null? res) '() (html:br)) ,res))
        '()
        (reverse (string-split (html-aa-escape-string str) #\lf))))

(define (textize article)
  (string-append (header-of article)
                 "\n"
                 (body-of article)
                 "\n"
                 (footer-of article)))


;; part writer

(define (common-header . title)
  (html:head
   (html:meta :http-equiv "content-type"
              :content #`"text/html; charset=,(cgi-output-character-encoding)")
   (if (robots-noindex?)
       (html:meta :name "robots" :content "noindex,follow")
       '())
   (html:title (if (not (null? title))
                   title
                   (page-title)))
   (if (page-keywords)
       (html:meta :name "keywords" :content (page-keywords))
       '())
   (if (and (null? title) (page-description))
       (html:meta :name "description" :content (page-description))
       '())
   (html:link :rel "start" :target "_top" :href (cgi-path "c" "f"))
   (html:link :rel "contents" :href (cgi-path "c" "tl"))
   (html:link :rel "index" :href (cgi-path "c" "i"))
   (html:link :rel "alternate" :title "Atom 1.0"
              :href (cgi-path "c" "ra") :type "application/atom+xml")
   (map (compose (pa$ html:link :rel "stylesheet" :type "text/css" :href)
                 fix-path)
        (page-css))
   (map (compose (pa$ html:script :type "text/javascript" :src)
                 fix-path)
        (page-js))
   ))

(define (cleantext-script)
  (html:script
   :type "text/javascript"
   "function splitheader() {"
   "  var body = document.getElementById('body');"
   "  var header = document.getElementById('header');"
   "  var m = body.value.split('\\n');"
   "  if(header.value == '') {"
   "    var str = m[0].substring(0, m[0].length - 1);"
   "    str = str.replace(/^[0-9]* /, '');"
   "    str = str.replace("
   "/([^0-9a-zA-Z])[0-9a-zA-Z-.]+\\.(jp|com|net)([^0-9a-zA-Z])/"
   "                      , '$1$3');"
   "    str = str.replace(/ \\?[^ ]*$/, '');"
   "    str = str.replace(/ [^ 0-9] ID:/, ' ID:');"
   "    header.value = str;"
   "  } else {"
   "    header.value = header.value + '\\n' + "
   "                   m[0].substring(0, m[0].length - 1);"
   "  }"
   "  m.shift();"
   "  body.value = m.join('\\n');"
   "}"
   "function splitfooter() {"
   "  var body = document.getElementById('body');"
   "  var footer = document.getElementById('footer');"
   "  var m = body.value.split('\\n');"
   "  if(footer.value == '') {"
   "    footer.value = m.pop();"
   "  } else {"
   "    footer.value = m.pop() + '\\n' + footer.value;"
   "  }"
   "  m = m.join('\\n');"
   "  body.value = m.substring(0, m.length - 1);"
   "}"
   "function splitfooter_auto() {"
   "  var body = document.getElementById('body');"
   "  var footer = document.getElementById('footer');"
   "  var m = body.value.split('\\r\\n');"
   "  var preppendfooter = function() {"
   "    if(footer.value == '') {"
   "      footer.value = m.pop();"
   "    } else {"
   "      footer.value = m.pop() + '\\n' + footer.value;"
   "    }"
   "  };"
   "  preppendfooter();"
   "  while(m.length > 0 && m[m.length - 1] == '') {"
   "    preppendfooter();"
   "  }"
   "  body.value = m.join('\\n');"
   "}"
   "function cleanbody() {"
   "  var body = document.getElementById('body');"
   "  var str = body.value;"
   "  str = str.replace(/[ \u3000]+$/mg, '');"
   "  str = str.replace(/\\r/g, '');"
   "  var m = str.split('\\n');"
   "  while(m.length > 0 && m[m.length - 1] == '') { m.pop(); }"
   "  str = m.join('\\n');"
   "  body.value = str;"
   "}"
   "function pastebody() {"
   "  var body = document.getElementById('body');"
   "  var str = window.clipboardData.getData('Text');"
   "  body.value = str;"
   "}"
   "function paste_clean_split() {"
   "  pastebody();"
   "  cleanbody();"
   "  splitheader();"
   "  splitfooter_auto();"
   "}"
   ))

(define (slist-part title-id title separator target)
  (html:p
   (map (lambda (subtitle-id)
          (let ((subtitle (get-subtitle title subtitle-id)))
            (list
             (html:span
              :id #`"s-,|subtitle-id|"
              (html:a :target target
                      :href (cgi-path "c" "al"
                                      "t" title-id
                                      "s" subtitle-id)
                      (name-of subtitle))
              "(" (length (articles-of subtitle)) ")"
              )
             separator
             )))
        (subtitles-of title))
   (if (related-page-of title)
       (html:a :target target
               :href (cgi-path "c" "tp"
                               "t" title-id)
               :id "tp"
               "関連ページ")
       '())
   (if (page-editable?)
       `(,separator
         ,(html:a :target target
                  :href (cgi-path "c" "se"
                                  "t" title-id)
                  "Edit"))
       '())
   ))

(define (aas-link:title title-id)
  (let* ((root (get-root-data))
         (title (get-title root title-id))
         )
    (html:a :target "content"
            :href (cgi-path "c" "sf"
                            "t" title-id)
            "項目:"
            (name-of title))))

(define (aas-link:subtitle title-id subtitle-id)
  (let* ((root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (html:a :target "content"
            :href (cgi-path "c" "sf"
                            "t" title-id
                            "s" subtitle-id)
            "項目:"
            #`",(name-of title) / ,(name-of subtitle)")))

(define (aas-link:blank url node)
  (html:a :target "_blank" :href url node))

(define (aas-link:wikipedia name)
  (let ((ename (uri-encode-string
                (ces-convert name (gauche-character-encoding) "utf-8")
                :noescape #[-_.!~*'()0-9A-Za-z/])))
    (aas-link:blank #`"http://ja.wikipedia.org/wiki/,|ename|"
                    #`"Wikipedia:,|name|")))


;; top frame writer

(define (frame-writer)
  `(,(html-doctype :type :html-4.01-frameset)
    ,(html:html
      :class "with-frame"
      (common-header)
      (html:script
       :type "text/javascript"
       (string-join
        '(
          "var loadCompleted = false;"
          "var onloadcompleted = null;"
          "window.onload = function() { "
          "  loadCompleted = true;"
          "  if(onloadcompleted) { onloadcompleted(); }"
          "}"
          ""
          "var selectedElem = null;"
          "var orgClassName = '';"
          "function highlight_t(name) {"
          "  if(! loadCompleted) {"
          "    onloadcompleted = function() {"
          "      onloadcompleted = null;"
          "      highlight_t(name);"
          "    };"
          "  }"
          "  if(selectedElem) {"
          "    selectedElem.className = orgClassName;"
          "    selectedElem = null;"
          "  }"
          "  var menu = document.getElementsByName('menu')[0];"
          "  var tlink = menu.contentWindow.document.getElementById(name);"
          "  if(tlink) {"
          "    selectedElem = tlink;"
          "    orgClassName = selectedElem.className;"
          "    selectedElem.className = orgClassName + ' selected';"
          "  }"
          "}"
          )
        "\n"))
      (html:frameset
       :class "top"
       :cols "200,*"
       (html:frame :name "menu"
                   :src (cgi-path "c" "tl"))
       (html:frame :name "content"
                   :src (cgi-path "c" "t"))
       (html:noframes
        (html:body
         (html:h1 :class "page-title" (page-title))
         (html:a :href (cgi-path "c" "tl") "Title menu")
         (html:br)
         (html:a :href (cgi-path "c" "t") "Top page")
         ))
       ))))

(define (top-writer)
  ((top-page-writer)
   (let ((root (get-root-data)))
     (apply + (map (lambda (title)
                     (apply + (map (lambda (subtitle)
                                     (length (articles-of subtitle)))
                                   (map (pa$ get-subtitle title)
                                        (subtitles-of title)))))
                   (map (pa$ get-title root) (titles-of root)))))))

(define (default-top-page-writer total-articles)
  `(,(html-doctype :type :html-4.01-strict)
    ,(html:html
      :class "top"
      (common-header "top")
      (html:body
       (html:p
        (html:h1 :class "page-title" (page-title))
        "Total # of AAs: " total-articles
        )
       (html:p
        (html:a :target "_self"
                :href (cgi-path "c" "r")
                "Recent changes"))
       ))))

(define (index-writer)
  `(,(html-doctype :type :html-4.01-strict)
    ,(html:html
      (parameterize ((robots-noindex? #t))
        (common-header "index"))
      (html:body
       :class "index"
       (html:script
        :type "text/javascript"
        (string-join
         '("function prepare()"
           "{"
           "    var ts = document.getElementsByTagName('li');"
           "    var lis = new Array();"
           "    var tn = 0;"
           "    var re = /<[^>]*>/g;"
           "    for(var i = 0; i < ts.length; i++) {"
           "        lis[i * 2] = ts[i].innerHTML.replace(re, '').toLowerCase();"
           "        lis[i * 2 + 1] = ts[i];"
           "    }"
           ""
           "    var p = document.createElement('p');"
           "    p.className = 'search';"
           ""
           "    p.appendChild(document.createTextNode('絞り込み検索: '));"
           ""
           "    var ti = document.createElement('input');"
           "    ti.type = 'text';"
           "    ti.id = 'txt';"
           "    ti.className = 'found';"
           "    ti.value = '';"
           "    ti.onkeyup = searcher(lis, 'list-item', ti.id);"
           "    p.appendChild(ti);"
           ""
           "    var hn = document.getElementsByTagName('h2')[0].nextSibling;"
           "    hn.parentNode.insertBefore(p, hn);"
           "}"
           "function searcher(items, display, id)"
           "{"
           "    var prev_txt = '';"
           "    function search()"
           "    {"
           "        var t = document.getElementById(id);"
           "        var txt = t.value.toLowerCase();"
           "        if(txt == prev_txt) {"
           "            return;"
           "        }"
           ""
           "        var found = false;"
           ""
           "        for(var i = 0; i < items.length / 2; i++) {"
           "            if(txt.length == 0 ||"
           "               items[i * 2].indexOf(txt) != -1) {"
           "                items[i * 2 + 1].style.display = display;"
           "                found = true;"
           "            } else {"
           "                items[i * 2 + 1].style.display = 'none';"
           "            }"
           "        }"
           ""
           "        if(found) {"
           "            t.className = 'found';"
           "        } else {"
           "            t.className = 'notfound';"
           "        }"
           ""
           "        prev_txt = txt;"
           "    }"
           ""
           "    var timer_id = false;"
           "    function ontimeout()"
           "    {"
           "        timer_id = false;"
           "        search();"
           "    }"
           ""
           "    function onkeyup()"
           "    {"
           "        if(timer_id) {"
           "            window.clearTimeout(timer_id);"
           "            timer_id = false;"
           "        }"
           "        timer_id = window.setTimeout(ontimeout, 200);"
           "    }"
           ""
           "    return onkeyup;"
           "}"
           ""
           "function highlight() {"
           "  if(window.parent && window.parent.highlight_t) {"
           "    window.parent.highlight_t('index');"
           "  }"
           "}"
           ""
           "window.onload = function() {"
           "  prepare();"
           "  highlight();"
           "}"
           )
         "\n"))
       (html:h2 "Index")
       (let ((root (get-root-data))
             (name-cmp (lambda (x y) (string-jpci<? (name-of x) (name-of y))))
             )
         (html:ul
          (map (lambda (title)
                 (let ((subtitles (map (pa$ get-subtitle title)
                                       (subtitles-of title))))
                   (html:li
                    (html:a :target "_self"
                            :href (cgi-path "c" "sf"
                                            "t" (base-id-of title))
                            (name-of title))
                    " ("
                    (apply + (map (compose length articles-of) subtitles))
                    ")"
                    (html:ul
                     (map (lambda (subtitle)
                            (html:li
                             (html:a :target "_self"
                                     :href (cgi-path "c" "sf"
                                                     "t" (base-id-of title)
                                                     "s" (base-id-of subtitle))
                                     (name-of subtitle))
                             " ("
                             (length (articles-of subtitle))
                             ")"))
                          (sort subtitles name-cmp))))))
               (sort (map (pa$ get-title root) (titles-of root)) name-cmp))))
       (html:hr)
       (html:a :target "_top" :href (cgi-path "c" "f") "Top")
       ))))


;; title writers

(define (tlist-writer)
  (%tlist-writer #f))

(define (tlist-sorted-writer)
  (%tlist-writer #t))

(define (%tlist-writer sort?)
  `(,(html-doctype :type :html-4.01-strict)
    ,(html:html
      (parameterize ((robots-noindex? #t))
        (common-header "menu"))
      (html:body
       :class "tlist"
       (html:h2 :class "page-title" (page-title))
       (if (page-editable?)
           (html:p
            (html:a :target "content"
                    :href (cgi-path "c" "te")
                    "Edit"))
           '())
       ((pre-tlist-proc))
       (html:div
        :class "sort-navi"
        "Sort: "
        (if sort?
            (html:span :class "btn-c" "ON")
            (html:a :href (cgi-path "c" "tls") :class "btn" "ON"))
        " "
        (if sort?
            (html:a :href (cgi-path "c" "tl") :class "btn" "OFF")
            (html:span :class "btn-c" "OFF"))
        )
       (let ((root (get-root-data))
             (s$ (if sort?
                     (lambda (l)
                       (sort l (lambda (x y)
                                 (string-jpci<? (name-of x) (name-of y)))))
                     values)))
         (map (lambda (title)
                (html:div
                 :class "tname"
                 :id #`"t-,(base-id-of title)"
                 (html:a :target "content"
                         :href (cgi-path "c" "sf"
                                         "t" (base-id-of title))
                         (name-of title))
                 " ("
                 (apply + (map (compose length articles-of)
                               (map (pa$ get-subtitle title)
                                    (subtitles-of title))))
                 ")"
                 ))
              (s$ (map (pa$ get-title root) (titles-of root)))))
       (html:p
        :id "rchange"
        (html:a :target "content"
                :href (cgi-path "c" "r")
                "Recent changes"))
       (html:p
        :id "index"
        (html:a :target "content"
                :href (cgi-path "c" "i")
                "Index"))
       (html:p
        :id "top"
        (html:a :target "_top"
                :href (cgi-path "c" "f")
                "Top"))
       (if (page-editable?)
           (html:p
            (html:a :target "content"
                    :href (cgi-path "c" "ex")
                    "Export"))
           '())
       ((post-tlist-proc))
       ))))

(define (tedit-writer)
  (let* ((root (get-root-data))
         (title-id-names (map (lambda (title-id)
                                (let ((title (get-title root title-id)))
                                  (cons title-id (name-of title))))
                              (titles-of root)))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`"edit titles")
        (html:body
         :class "tedit"
         :onload "fill_edit_name();"
         (html:script
          :type "text/javascript"
          "function fill_edit_name() {"
          "  var elem = document.getElementById('edit-sel');"
          "  document.getElementById('edit-name').value = "
          "    elem.options(elem.selectedIndex).text;"
          "  document.getElementById('edit-id').value = "
          "    elem.options(elem.selectedIndex).value;"
          "}"
          )
         (html:p
          "Edit title name"
          (html:br)
          (html:form
           :method "POST"
           :target "_top"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "tu")
           (html:select
            :name "t"
            :onchange "fill_edit_name();"
            :id "edit-sel"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 (sort title-id-names
                       (lambda (x y)
                         (string-jpci<? (cdr x) (cdr y))))))
           (html:br)
           (html:input :type "text" :name "n" :value "" :id "edit-name")
           (html:input :type "text" :name "ni" :value "" :id "edit-id")
           (html:input :type "submit" :value "Change")
          ))
         (html:hr)
         (html:p
          "Move title pos"
          (html:br)
          (html:form
           :method "POST"
           :target "_top"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "tm")
           "move "
           (html:select
            :name "t"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 title-id-names))
           (html:br)
           "before "
           (html:select
            :name "n"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 title-id-names)
            (html:option :value "" "(Last)"))
           (html:br)
           (html:input :type "submit" :value "Move")
          ))
         (html:hr)
         (html:p
          "Add title"
          (html:br)
          (html:form
           :method "POST"
           :target "_top"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "ta")
           "ID: "
           (html:input :type "text" :name "t" :value "")
           (html:br)
           "Name: "
           (html:input :type "text" :name "n" :value "")
           (html:br)
           (html:input :type "submit" :value "Add")
           ))
         (html:hr)
         (html:p
          "Remove title"
          (html:br)
          (html:form
           :method "POST"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "trc")
           (html:select
            :name "t"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 title-id-names))
           (html:br)
           (html:input :type "submit" :value "Remove")
          ))
         (html:hr)
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         )))))

(define (tupdate-writer)
  (let* ((title-id (get-title-id))
         (name (get-edit-name))
         (id (get-edit-id))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (when (and (not (equal? title-id id))
               (get-title root id))
      (error "specified title-id is already exists" id))
    (unless (equal? (name-of title) name)
      (update-title! title name))
    (unless (equal? title-id id)
      (update-title-id! root title-id id))
    (success-writer (page-path))))

(define (tmove-writer)
  (let* ((title-id (get-title-id))
         (next-id (get-edit-name))
         (root (get-root-data))
         )
    (move-title! root title-id next-id)
    (success-writer (page-path))))

(define (tadd-writer)
  (let* ((title-id (get-title-id))
         (name (get-edit-name))
         (root (get-root-data))
         )
    (when (get-title root title-id)
      (error "specified title-id is already exists" title-id))
    (unless (#/^[0-9a-zA-Z-_]+$/ title-id)
      (error "invalid title-id" title-id))
    (add-title! root title-id name)
    (success-writer (page-path))))

(define (trm-confirm-writer)
  (let* ((title-id (get-title-id))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`"confirm to remove ,(name-of title)")
        (html:body
         :class "trm-confirm"
         (html:p
          #`"Removing these subtitles of ,(name-of title)"
          (html:ul
           (map (lambda (subtitles-id)
                  (let ((subtitle (get-subtitle title subtitles-id)))
                    (html:li
                     #`",(name-of subtitle) (,(length (articles-of subtitle)))"
                     )))
                (subtitles-of title)))
          (html:form
           :method "POST"
           :target "_top"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "tr")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:input :type "submit" :value "Remove")
           ))
         )))))

(define (tremove-writer)
  (let* ((title-id (get-title-id))
         (root (get-root-data))
         )
    (remove-title! root title-id)
    (success-writer (page-path))))

(define (trelpage-writer)
  (let* ((title-id (get-title-id))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`",(name-of title) / Related page")
        (html:body
         :class "relpage"
         (html:script
          :type "text/javascript"
          #`"var tname = 't-,|title-id|';\n"
          "var sname = 'tp';\n"
          (string-join
           '(
             "window.onload = function() {"
             "  if(window.parent && window.parent.parent &&"
             "     window.parent.parent.highlight_t) {"
             "    window.parent.parent.highlight_t(tname);"
             "  }"
             "  if(window.parent && window.parent.highlight_t) {"
             "    window.parent.highlight_t(tname);"
             "  }"
             "  if(window.parent && window.parent.highlight_s) {"
             "    window.parent.highlight_s(sname);"
             "  }"
             "}"
             )
           "\n")
          )
         (if (related-page-of title)
             (map (lambda (sexp) (eval sexp (current-module)))
                  (call-with-input-string (related-page-of title)
                    port->sexp-list))
             '())
         (html:hr)
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         " / "
         (html:a :target "content" :href (cgi-path "c" "sf"
                                                   "t" title-id)
                 "Title frame")
         )))))

(define (trelupdate-writer)
  (let* ((title-id (get-title-id))
         (relpage (get-relpage-text))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (update-title-relpage! title (and (not (string=? relpage ""))
                                      relpage))
    (success-writer (cgi-path "c" "tp"
                              "t" title-id))))

(define (tkeywords-writer)
  (let* ((title-id (get-title-id))
         (keywords (get-keywords-text))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (update-title-keywords! title (and (not (string=? keywords ""))
                                       keywords))
    (success-writer (cgi-path "c" "se"
                              "t" title-id))))


;; subtitle writers

(define (sframe-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id*))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    `(,(html-doctype :type :html-4.01-frameset)
      ,(html:html
        :class "with-frame"
        (common-header (name-of title))
        (html:script
         :type "text/javascript"
         (string-join
          '(
            "function fit_size() {"
            "  var sframe = document.getElementById('sframe');"
            "  var smenu = document.getElementsByName('smenu')[0];"
            "  var sframe_h ="
            "    window.innerHeight ?"
            "    window.innerHeight :"
            "    document.documentElement.scrollHeight;"
            "  var smenu_h ="
            "    smenu.contentWindow.innerHeight ?"
            "    smenu.contentWindow.document.documentElement.offsetHeight :"
            "    smenu.contentWindow.document.documentElement.scrollHeight;"
            "  if(smenu_h * 3 > sframe_h) {"
            "    smenu_h = sframe_h / 3;"
            "  }"
            "  sframe.setAttribute('rows', smenu_h + ',*');"
            "}"
            "window.onresize = fit_size;"
            ""
            "var loadCompleted = false;"
            "var onloadcompleted = null;"
            "window.onload = function() { "
            "  loadCompleted = true;"
            "  if(onloadcompleted) { onloadcompleted(); }"
            "  fit_size();"
            "}"
            ""
            "var selectedElem = null;"
            "var orgClassName = '';"
            "function highlight_s(name) {"
            "  if(! loadCompleted) {"
            "    onloadcompleted = function() {"
            "      onloadcompleted = null;"
            "      highlight_s(name);"
            "    };"
            "  }"
            "  if(selectedElem) {"
            "    selectedElem.className = orgClassName;"
            "    selectedElem = null;"
            "  }"
            "  var menu = document.getElementsByName('smenu')[0];"
            "  var slink = menu.contentWindow.document.getElementById(name);"
            "  if(slink) {"
            "    selectedElem = slink;"
            "    orgClassName = selectedElem.className;"
            "    selectedElem.className = orgClassName + ' selected';"
            "  }"
            "}"
            )
          "\n"))
        (html:frameset
         :class "sframe"
         :id "sframe"
         :rows "80,*"
         (html:frame :name "smenu"
                     :src (cgi-path "c" "sl"
                                    "t" title-id))
         (html:frame :name "aa"
                     :src (if (null? (subtitles-of title))
                              "about:blank"
                              (cgi-path "c" "al"
                                        "t" title-id
                                        "s" (if subtitle-id
                                                subtitle-id
                                                (car (subtitles-of title))))))
         (html:noframes
          (html:body
           (html:h2 (name-of title))
           (if subtitle-id
               (html:a :href (cgi-path "c" "sf"
                                       "t" title-id)
                       "submenu")
               (slist-part title-id title (html:br) #f))))
         )))))

(define (slist-writer)
  (let* ((title-id (get-title-id))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`"submenu of ,(name-of title)")
        (html:body
         :class "smenu"
         (slist-part title-id title " " "aa")
         )))))

(define (sedit-writer)
  (let* ((title-id (get-title-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle-id-names
          (map (lambda (subtitle-id)
                 (let ((subtitle (get-subtitle title subtitle-id)))
                   (cons subtitle-id (name-of subtitle))))
               (subtitles-of title)))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`"edit subtitles")
        (html:body
         :class "sedit"
         :onload "fill_edit_name();"
         (html:script
          :type "text/javascript"
          "function fill_edit_name() {"
          "  var elem = document.getElementById('edit-sel');"
          "  document.getElementById('edit-name').value = "
          "    elem.options(elem.selectedIndex).text;"
          "  document.getElementById('edit-id').value = "
          "    elem.options(elem.selectedIndex).value;"
          "}"
          )
         (html:p
          "Edit subtitle name"
          (html:br)
          (html:form
           :method "POST"
           :target "content"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "su")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:select
            :name "s"
            :onchange "fill_edit_name();"
            :id "edit-sel"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 subtitle-id-names))
           (html:br)
           (html:input :type "text" :name "n" :value "" :id "edit-name")
           (html:input :type "text" :name "ni" :value "" :id "edit-id")
           (html:input :type "submit" :value "Change")
          ))
         (html:hr)
         (html:p
          "Move subtitle pos"
          (html:br)
          (html:form
           :method "POST"
           :target "content"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "sm")
           (html:input :type "hidden" :name "t" :value title-id)
           "move "
           (html:select
            :name "s"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 subtitle-id-names))
           (html:br)
           "before "
           (html:select
            :name "n"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 subtitle-id-names)
            (html:option :value "" "(Last)"))
           (html:br)
           (html:input :type "submit" :value "Move")
          ))
         (html:hr)
         (html:p
          "Add subtitle"
          (html:br)
          (html:form
           :method "POST"
           :target "content"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "sa")
           (html:input :type "hidden" :name "t" :value title-id)
           "ID: "
           (html:input :type "text" :name "s" :value "")
           (html:br)
           "Name: "
           (html:input :type "text" :name "n" :value "")
           (html:br)
           (html:input :type "submit" :value "Add")
           ))
         (html:hr)
         (html:p
          "Remove subtitle"
          (html:br)
          (html:form
           :method "POST"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "src")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:select
            :name "s"
            (map (lambda (id-name)
                   (html:option :value (car id-name) (cdr id-name)))
                 subtitle-id-names))
           (html:br)
           (html:input :type "submit" :value "Remove")
          ))
         (html:hr)
         (html:p
          "Related page"
          (html:br)
          (html:form
           :method "POST"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "tpu")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:textarea :class "script" :name "r"
                          (textarea-escape-string
                           (or (related-page-of title)
                               (string-join
                                '("(html:h2 \"関連ページ\")"
                                  ""
                                  )
                                "\n"))))
           (html:input :type "submit" :value "Update")
           " " (html:input :type "reset" :value "Reset")
           ))
         (html:hr)
         (html:p
          "Keywords"
          (html:br)
          (html:form
           :method "POST"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "tk")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:input :type "text" :name "k" :value (keywords-of title))
           (html:input :type "submit" :value "Update")
           " " (html:input :type "reset" :value "Reset")
           ))
         (html:hr)
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         )))))

(define (supdate-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (name (get-edit-name))
         (id (get-edit-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (when (and (not (equal? subtitle-id id))
               (get-subtitle title id))
      (error "specified subtitle-id is already exists" id))
    (unless (equal? (name-of subtitle) name)
      (update-subtitle! subtitle name))
    (unless (equal? subtitle-id id)
      (update-subtitle-id! title subtitle-id id))
    (success-writer (cgi-path "c" "sf"
                              "t" title-id
                              "s" id))))

(define (smove-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (next-id (get-edit-name))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (move-subtitle! title subtitle-id next-id)
    (success-writer (cgi-path "c" "sf"
                              "t" title-id
                              "s" subtitle-id))))

(define (sadd-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (name (get-edit-name))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (when (get-subtitle title subtitle-id)
      (error "specified subtitle-id is already exists" subtitle-id))
    (unless (#/^[0-9a-zA-Z-_]+$/ subtitle-id)
      (error "invalid subtitle-id" subtitle-id))
    (add-subtitle! title subtitle-id name)
    (success-writer (cgi-path "c" "sf"
                              "t" title-id
                              "s" subtitle-id))))

(define (srm-confirm-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header
         #`"confirm to remove ,(name-of title) / ,(name-of subtitle)")
        (html:body
         :class "srm-confirm"
         (html:p
          #`"Removing ,(length (articles-of subtitle)) articles: "
          #`",(name-of title) / ,(name-of subtitle)"
          (html:form
           :method "POST"
           :target "content"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "sr")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:input :type "hidden" :name "s" :value subtitle-id)
           (html:input :type "submit" :value "Remove")
           ))
         )))))

(define (sremove-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (root (get-root-data))
         (title (get-title root title-id))
         )
    (remove-subtitle! title subtitle-id)
    (success-writer (cgi-path "c" "sf"
                              "t" title-id))))

(define (sheading-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header
         #`"heading of ,(name-of title) / ,(name-of subtitle)")
        (html:body
         :class "sheading"
         (html:h2 (name-of subtitle))
         (html:p "Heading")
         (html:form
          :method "POST"
          :action (page-path)
          (html:input :type "hidden" :name "c" :value "shu")
          (html:input :type "hidden" :name "t" :value title-id)
          (html:input :type "hidden" :name "s" :value subtitle-id)
          (html:textarea :class "heading" :name "h" (heading-of subtitle))
          (html:textarea :class "script" :name "r" (script-of subtitle))
          (html:input :type "submit" :value "Update")
          " " (html:input :type "reset" :value "Reset")
          )
         (html:hr)
         (html:a :target "aa"
                 :href (cgi-path "c" "al"
                                 "t" title-id
                                 "s" subtitle-id)
                 "Return")
         )))))

(define (shupdate-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (heading (get-heading-text))
         (script (get-script-text))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (update-subtitle-heading! subtitle heading script)
    (success-writer (cgi-path "c" "al"
                              "t" title-id
                              "s" subtitle-id))))


;; article writers

(define (alist-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         (heading (if (string=? (heading-of subtitle) "")
                      (name-of subtitle)
                      (heading-of subtitle)))
         (proc-args (list title-id subtitle-id
                          (name-of title) (name-of subtitle)
                          (keywords-of title)
                          (length (subtitles-of title))
                          (length (articles-of subtitle))
                          ))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header #`",(name-of title) / ,heading")
        (html:body
         :class "alist"
         (html:script
          :type "text/javascript"
          #`"var tname = 't-,|title-id|';\n"
          #`"var sname = 's-,|subtitle-id|';\n"
          (string-join
           '(
             "window.onload = function() {"
             "  highlight();"
             "  if(window.init) { init(); }"
             "}"
             ""
             "function highlight() {"
             "  if(window.parent && window.parent.parent &&"
             "     window.parent.parent.highlight_t) {"
             "    window.parent.parent.highlight_t(tname);"
             "  }"
             "  if(window.parent && window.parent.highlight_t) {"
             "    window.parent.highlight_t(tname);"
             "  }"
             "  if(window.parent && window.parent.highlight_s) {"
             "    window.parent.highlight_s(sname);"
             "  }"
             "}"
             ""
             "function copyaa(str) {"
             "  window.clipboardData.setData('Text', str);"
             "}"
             )
           "\n")
          )
         (html:h2 heading)
         (if (string=? (script-of subtitle) "")
             '()
             (script-of subtitle))
         (if (page-editable?)
             `(,(html:a :target "aa"
                        :href (cgi-path "c" "an"
                                        "t" title-id
                                        "s" subtitle-id)
                        "Add")
               " "
               ,(html:a :target "aa"
                        :href (cgi-path "c" "sh"
                                        "t" title-id
                                        "s" subtitle-id)
                        "Heading")
               ,(html:hr))
             '())
         (apply (pre-alist-proc) proc-args)
         (map (lambda (article-id)
                (let ((article (get-article subtitle article-id)))
                  `(,(html:div
                      :id #`"aa,|article-id|"
                      (html:a
                       :name article-id
                       (html:div :class "aa" (htmlize (header-of article))))
                      (html:div :class (style-of article)
                                :id #`"aab,|article-id|"
                                (htmlize (body-of article)))
                      (html:div :class "aa" (htmlize (footer-of article)))
                      (html:script
                       :type "text/javascript"
                       "var aasrc" article-id " = \""
                       (jscript-escape-string (body-of article))
                       "\";"
                       "if(window.clipboardData) {"
                       "document.write('"
                       "<p class=\"aa-op\">"
                       "<a href=\"javascript:copyaa(aasrc"
                       article-id
                       ");\">Copy</a></p>"
                       "')"
                       "}"
                       )
                      (if (page-editable?)
                          (html:div
                           (html:a :target "aa"
                                   :href (cgi-path "c" "ae"
                                                   "t" title-id
                                                   "s" subtitle-id
                                                   "a" article-id)
                                   "Edit")
                           " "
                           (html:a :target "aa"
                                   :href (cgi-path "c" "arc"
                                                   "t" title-id
                                                   "s" subtitle-id
                                                   "a" article-id)
                                   "Remove")
                           )
                          '()))
                    ,(html:hr)
                    )))
              (articles-of subtitle))
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         " / "
         (html:a :target "content" :href (cgi-path "c" "sf"
                                                   "t" title-id
                                                   "s" subtitle-id)
                 "Title frame")
         (apply (post-alist-proc) proc-args)
         )))))

(define (aedit-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (article-id (get-article-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         (article (get-article subtitle article-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header
         #`"edit ,(name-of title) / ,(name-of subtitle) / ,|article-id|")
        (html:body
         :class "aedit"
         (cleantext-script)
         (html:h2 (name-of subtitle))
         (html:p "Edit")
         (html:form
          :method "POST"
          :action (page-path)
          (html:input :type "hidden" :name "c" :value "au")
          (html:input :type "hidden" :name "t" :value title-id)
          (html:input :type "hidden" :name "s" :value subtitle-id)
          (html:input :type "hidden" :name "a" :value article-id)
          (html:textarea :class "header" :id "header"
                         :name "h"
                         (textarea-escape-string (header-of article)))
          (html:br)
          (html:textarea :class "body" :id "body"
                         :name "b"
                         (textarea-escape-string (body-of article)))
          (html:br)
          (html:textarea :class "footer" :id "footer"
                         :name "f"
                         (textarea-escape-string (footer-of article)))
          (html:p
           :class "aa-op"
           (html:input :type "submit" :value "Change")
           (html:input :type "button" :onclick "paste_clean_split()"
                                      :value "Paste, Clean and Split")
           " " (html:select
                :name "st"
                (map (lambda (style)
                       (html:option :value style
                                    :selected (string=? (style-of article)
                                                        style)
                                    style))
                     (map car (aa-style-alist))))
           " " (html:input :type "reset" :value "Reset")
           (html:br)
           (html:input :type "button" :onclick "cleanbody()"
                                      :value "Clean Body")
           (html:input :type "button" :onclick "splitheader()"
                                      :value "Split Header")
           (html:input :type "button" :onclick "splitfooter()"
                                      :value "Split Footer")
           (html:input :type "button" :onclick "splitfooter_auto()"
                                      :value "Split Footer (auto)")
           )
          (html:table
           (html:tr (html:td "Registered: ")
                    (html:td (date->string (c-date-of article) date-format)))
           (html:tr (html:td "Last modified: ")
                    (html:td (date->string (m-date-of article) date-format)))
           )
          )
         (html:hr)
         (html:a :target "aa"
                 :href (cgi-path "c" "al"
                                 "t" title-id
                                 "s" subtitle-id)
                 "Return")
         )))))

(define (anew-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header
         #`"add ,(name-of title) / ,(name-of subtitle)")
        (html:body
         :class "anew"
         (cleantext-script)
         (html:h2 (name-of subtitle))
         (html:p "New")
         (html:form
          :method "POST"
          :action (page-path)
          (html:input :type "hidden" :name "c" :value "aa")
          (html:input :type "hidden" :name "t" :value title-id)
          (html:input :type "hidden" :name "s" :value subtitle-id)
          (html:textarea :class "header" :id "header" :name "h")
          (html:br)
          (html:textarea :class "body" :id "body" :name "b")
          (html:br)
          (html:textarea :class "footer" :id "footer" :name "f")
          (html:p
           :class "aa-op"
           (html:input :type "submit" :value "Add")
           (html:input :type "button" :onclick "paste_clean_split()"
                                      :value "Paste, Clean and Split")
           " " (html:select
                :name "st"
                (map (lambda (style)
                       (html:option :value style
                                    :selected (string=? (aa-default-style)
                                                        style)
                                    style))
                     (map car (aa-style-alist))))
           " " (html:input :type "reset" :value "Reset")
           (html:br)
           (html:input :type "button" :onclick "cleanbody()"
                                      :value "Clean Body")
           (html:input :type "button" :onclick "splitheader()"
                                      :value "Split Header")
           (html:input :type "button" :onclick "splitfooter()"
                                      :value "Split Footer")
           (html:input :type "button" :onclick "splitfooter_auto()"
                                      :value "Split Footer (auto)")
           )
          )
         (html:hr)
         (html:a :target "aa"
                 :href (cgi-path "c" "al"
                                 "t" title-id
                                 "s" subtitle-id)
                 "Return")
         )))))

(define (aupdate-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (article-id (get-article-id))
         (header (get-header-text))
         (body (get-body-text))
         (footer (get-footer-text))
         (style (get-style-text))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         (article (get-article subtitle article-id))
         )
    (update-article! article header footer body style)
    (success-writer (cgi-path "c" "al"
                              "t" title-id
                              "s" subtitle-id))))

(define (aadd-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (header (get-header-text))
         (body (get-body-text))
         (footer (get-footer-text))
         (style (get-style-text))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (add-article! subtitle header footer body style)
    (success-writer (cgi-path "c" "al"
                              "t" title-id
                              "s" subtitle-id))))

(define (arm-confirm-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (article-id (get-article-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         (article (get-article subtitle article-id))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header
         "confirm to remove "
         #`",(name-of title) / ,(name-of subtitle) / ,|article-id|")
        (html:body
         :class "arm-confirm"
         (html:p
          "Removing article "
          #`",(name-of title) / ,(name-of subtitle) / ,|article-id|")
         (html:p
          :class (style-of article)
          (htmlize (textize article)))
         (html:p
          (html:form
           :method "POST"
           :action (page-path)
           (html:input :type "hidden" :name "c" :value "ar")
           (html:input :type "hidden" :name "t" :value title-id)
           (html:input :type "hidden" :name "s" :value subtitle-id)
           (html:input :type "hidden" :name "a" :value article-id)
           (html:input :type "submit" :value "Remove")
           ))
         )))))

(define (aremove-writer)
  (let* ((title-id (get-title-id))
         (subtitle-id (get-subtitle-id))
         (article-id (get-article-id))
         (root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (remove-article! subtitle article-id)
    (success-writer (cgi-path "c" "al"
                              "t" title-id
                              "s" subtitle-id))))


;; recent changes writer

(define rchange-days (make-parameter #f))
(define rchange-feed-days (make-parameter #f))

(define (date->time-day-end time)
  (add-duration (date->time-utc time)
                (make-time time-duration
                           (- -1 (date-nanosecond time))
                           (- (* 60 60 24)
                              (+ (date-second time)
                                 (* 60
                                    (+ (date-minute time)
                                       (* 60 (date-hour time)))))
                              ))))

(define (rchange-writer)
  (let* ((root (get-root-data))
         (now (current-date))
         (today-end (date->time-day-end now))
         (since (subtract-duration today-end
                                   (make-time time-duration
                                              0 (* 60 60 24
                                                   (rchange-days)))))
         )
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header "recent changes")
        (html:body
         :class "rchange"
         (html:script
          :type "text/javascript"
          (string-join
           '(
             "window.onload = function() {"
             "  if(window.parent && window.parent.highlight_t) {"
             "    window.parent.highlight_t('rchange');"
             "  }"
             "}"
             )
           "\n")
          )
         (html:h2 "Recent changes")
         (html:div :class "feed-link"
                   (html:a :href (cgi-path "c" "ra")
                           "Atom 1.0"))
         (let ((rc
                ((compose
                  (cut stable-sort <>
                       (lambda (a1 a2)
                         (or (time>? (cadr a1) (cadr a2))
                             (and (time=? (cadr a1) (cadr a2))
                                  (eq? 'C (car a1))
                                  (eq? 'M (car a2)))
                             (and (time=? (cadr a1) (cadr a2))
                                  (eq? (car a1) (car a2))
                                  (string-jpci<? (caddr a1)
                                                 (caddr a2)))
                             )))
                  (remove$ null?)
                  (pa$
                   append-map
                   (lambda (article)
                     (let ((c-date
                            (date->time-day-end (c-date-of (car article))))
                           (m-date
                            (date->time-day-end (m-date-of (car article))))
                           (sid (list-ref article 3))
                           (tid (list-ref article 5))
                           (name
                            #`",(list-ref article 4) / ,(list-ref article 2)")
                           )
                       (list
                        (if (time>? c-date since)
                            (list 'C c-date name tid sid)
                            '())
                        (if (and (not (time=? c-date m-date))
                                 (time>? m-date since))
                            (list 'M m-date name tid sid)
                            '())))))
                  )
                 (rchange-articles since)))
               (counter (lambda (lst desc)
                          (let ((item (lambda (name ids cnt)
                                        (html:li
                                         desc
                                         (html:a :target "_self"
                                                 :href (cgi-path "c" "sf"
                                                                 "t" (car ids)
                                                                 "s" (cadr ids))
                                                 name)
                                         #`" (,|cnt|)"))))
                            (let loop ((lst lst)
                                       (name '())
                                       (ids '())
                                       (cnt 0)
                                       (res '()))
                              (cond ((null? lst)
                                     (reverse
                                      (if (> cnt 0)
                                          (cons (item name ids cnt) res)
                                          res)))
                                    ((equal? ids (cdddar lst))
                                     (loop (cdr lst)
                                           name
                                           ids
                                           (+ 1 cnt)
                                           res))
                                    (else
                                     (loop (cdr lst)
                                           (caddar lst)
                                           (cdddar lst)
                                           1
                                           (if (> cnt 0)
                                               (cons (item name ids cnt) res)
                                               res)))
                                     )))))
               )
           (let loop ((since (subtract-duration today-end
                                                (make-time time-duration
                                                           0 (* 60 60 24))))
                      (rc rc)
                      (res '()))
             (if (null? rc)
                 (reverse res)
                 (let*-values (((t o) (partition (lambda (r)
                                                   (time>? (cadr r) since))
                                                 rc))
                               ((c m) (partition (lambda (r) (eq? 'C (car r)))
                                                 t))
                               )
                   (loop (subtract-duration since (make-time time-duration
                                                             0 (* 60 60 24)))
                         o
                         (if (null? t)
                             res
                             (cons
                              (html:p
                               (date->string (time-utc->date (cadar t))
                                             "~Y/~m/~d")
                               (html:ul
                                (counter c "Add new: ")
                                (counter m "Modified: ")
                                ))
                              res)))))))
         (html:hr)
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         )))))

(define (xml tag attrs . contents)
  `("<" ,tag
    ,(map (lambda (attr)
            (list " " (car attr) "=\"" (cdr attr) "\""))
          attrs)
    ,(if (null? contents)
         " />\n"
         `(">" ,(map xml-escape contents) "</" ,tag ">\n"))
    ))

(define (xml-escape contents)
  (cond ((string? contents)
         (sxml:string->xml contents))
        (else
         contents)))

;; RFC 4287
(define (rchange-atom-writer)
  (let* ((format-time (lambda (time)
                        (date->string (time-utc->date time 0)
                                      "~Y-~m-~dT~H:~M:~SZ")))
         (format-date (lambda (date)
                        (format-time (date->time-utc date))))
         (since (subtract-duration (date->time-day-end (current-date))
                                   (make-time time-duration
                                              0 (* 60 60 24
                                                   (rchange-feed-days)))))
         (abs-path (lambda args
                     (parameterize ((page-base-path (page-base-url))
                                    (page-editable? #f))
                       (apply cgi-path args))))
         (entry (lambda (article tname sname tid sid aid)
                  (let ((uri #`",(abs-path \"c\" \"al\" \"t\" tid \"s\" sid)#,|aid|"))
                    `(,(date->time-utc (m-date-of article))
                      ,(xml 'entry '()
                            (xml 'category `((term ,tid)
                                             (label ,(sxml:string->xml tname))))
                            (xml 'title '((type "html"))
                                 #`",|tname| / ,|sname| / ,|aid|")
                            (xml 'link `((href ,(sxml:string->xml uri))
                                         (rel "alternate")
                                         (type "text/html")))
                            (xml 'id '() uri)
                            (xml 'published '()
                                 (format-date (c-date-of article)))
                            (xml 'updated '()
                                 (format-date (m-date-of article)))
                            (xml 'content '((type "html"))
                                 (call-with-output-string
                                   (pa$
                                    write-tree
                                    (list
                                     (html:div
                                      :style (assoc-ref (aa-style-alist) "aa")
                                      (htmlize (header-of article)))
                                     (html:div
                                      :style (assoc-ref (aa-style-alist)
                                                        (style-of article))
                                      (htmlize (body-of article)))
                                     (html:div
                                      :style (assoc-ref (aa-style-alist) "aa")
                                      (htmlize (footer-of article)))
                                     ))))
                            )
                      )
                    )))
         (entries ((compose
                    (cut stable-sort <>
                         (lambda (e1 e2) (time>? (car e1) (car e2))))
                    (remove$ null?)
                    (map$ (lambda (article)
                            (let ((tname (list-ref article 4))
                                  (sname (list-ref article 2))
                                  (tid (list-ref article 5))
                                  (sid (list-ref article 3))
                                  (aid (list-ref article 1))
                                  )
                              (entry (car article) tname sname tid sid aid))))
                    )
                   (rchange-articles since)))
         )
    (page-content-type "text/xml")
    `("<?xml version=\"1.0\" "
      ,#`"encoding=\",(cgi-output-character-encoding)\" ?>\n"
      ,(xml 'feed '((xmlns "http://www.w3.org/2005/Atom"))
            (xml 'title '((type "html"))
                 (page-title))
            (xml 'subtitle '((type "html"))
                 (page-description))
            (xml 'link `((href ,(sxml:string->xml (abs-path "c" "ra")))
                         (rel "self")
                         (type "application/atom+xml")))
            (xml 'link `((href ,(sxml:string->xml (abs-path "c" "f")))
                         (rel "alternate")
                         (type "text/html")))
            (xml 'author '()
                 (xml 'name '()
                      "aa.tamanegi.org"))
            (xml 'id '()
                 (abs-path "c" "f"))
            (xml 'updated '()
                 (format-time (caar entries)))
            (map cdr entries))
      )))

(define (rchange-articles since)
  (let ((root (get-root-data)))
    (append-map
     (lambda (title-id)
       (let* ((title (get-title root title-id))
              (title-name (name-of title)))
         (append-map
          (lambda (subtitle-id)
            (let* ((subtitle (get-subtitle title subtitle-id))
                   (subtitle-name (name-of subtitle)))
              (filter-map
               (lambda (article-id)
                 (let ((article (get-article subtitle article-id)))
                   (and (time>? (date->time-utc (m-date-of article)) since)
                        (list (get-article subtitle article-id)
                              article-id
                              subtitle-name
                              subtitle-id
                              title-name
                              title-id
                              ))))
               (articles-of subtitle))))
          (subtitles-of title))))
     (titles-of root))))


;; export writer

(define (export-writer)
  (let* ((write-to$
          (lambda (dest-dir . args)
            (lambda (fname writer desc)
              (let* ((fname (if (#/\/$/ fname)
                                #`",|fname|index.html"
                                fname))
                     (dest #`",|dest-dir|/,|fname|")
                     (dir (sys-dirname dest))
                     )
                (parameterize ((page-content-type (page-content-type))
                               (cgi-output-character-encoding
                                (cgi-output-character-encoding))
                               )
                  (let ((tree (writer)))
                    (make-directory* dir)
                    (if (apply call-with-output-file-if-changed
                               dest
                               (pa$ write-tree tree)
                               :encoding (cgi-output-character-encoding)
                               args)
                        `(,desc " (" ,dest ") ... wrote." ,(html:br))
                        '())))
                ))))
         (write-to (write-to$ (export-dest)))
         (write-est-to (if (est-export-dest)
                           (write-to$ (est-export-dest))
                           (lambda _ '())))
         )
    (make-directory* (export-dest))
    `(,(html-doctype :type :html-4.01-strict)
      ,(html:html
        (common-header "export result")
        (html:body
         :class "export"
         (html:p "Export result")
         (parameterize ((page-editable? #f))
           (html:div
            ((pre-export-proc))
            (let ((root (get-root-data)))
              (map
               (lambda (title-id)
                 (parameterize ((cgi-params (cons `("t" ,title-id)
                                                  (cgi-params))))
                   (let ((title (get-title root title-id)))
                     `(
                       ,(map
                         (lambda (subtitle-id)
                           (parameterize ((cgi-params (cons `("s" ,subtitle-id)
                                                            (cgi-params))))
                             `(
                               ,(write-to (cgi-path "c" "sf"
                                                    "t" title-id
                                                    "s" subtitle-id)
                                          sframe-writer
                                          `("subtitle frame: "
                                            ,title-id " / " ,subtitle-id))
                               ,(write-to (cgi-path "c" "al"
                                                    "t" title-id
                                                    "s" subtitle-id)
                                          alist-writer
                                          `("article list: "
                                            ,title-id " / " ,subtitle-id))

                               ,(write-est-to
                                 #`",|title-id|/,|subtitle-id|/"
                                 (pa$ est-s-export-writer title-id subtitle-id)
                                 `("est-s-export: "
                                   ,title-id " / " ,subtitle-id))

                               ,(let ((subtitle
                                       (get-subtitle title subtitle-id)))
                                  (map (lambda (article-id)
                                         (write-est-to
                                          #`",|title-id|/,|subtitle-id|/al#,|article-id|#.html"
                                          (pa$ est-a-export-writer
                                               title-id subtitle-id article-id)
                                          `("est-a-export: " ,title-id
                                            " / " ,subtitle-id
                                            " / " ,article-id)))
                                       (articles-of subtitle)))
                               )))
                         (subtitles-of title))

                       ,(if (related-page-of title)
                            (write-to (cgi-path "c" "tp"
                                                "t" title-id)
                                      trelpage-writer
                                      `("title rel page: " ,title-id))
                            '())

                       ,(write-to (cgi-path "c" "sl"
                                            "t" title-id)
                                  slist-writer
                                  `("subtitle list: " ,title-id))

                       ,(write-to (cgi-path "c" "sf"
                                            "t" title-id)
                                  sframe-writer
                                  `("subtitle frame: " ,title-id))

                       ,(write-est-to #`",|title-id|/"
                                      (pa$ est-t-export-writer title-id)
                                      `("est-t-export: " ,title-id))
                       ))))
               (titles-of root)))

            (write-to (cgi-path "c" "f")
                      frame-writer
                      "main frame")
            (write-to (cgi-path "c" "t")
                      top-writer
                      "top page")
            (write-to (cgi-path "c" "tl")
                      tlist-writer
                      "title list")
            (write-to (cgi-path "c" "tls")
                      tlist-sorted-writer
                      "title list (sorted)")
            (write-to (cgi-path "c" "r")
                      rchange-writer
                      "recent changes")
            (write-to (cgi-path "c" "ra")
                      rchange-atom-writer
                      "recent change Atom feed")
            (write-to (cgi-path "c" "i")
                      index-writer
                      "index")
            ((post-export-proc))
            ))
         (html:hr)
         (html:a :target "_top" :href (cgi-path "c" "f") "Top")
         )))))

(define (est-t-export-writer title-id)
  (let* ((root (get-root-data))
         (title (get-title root title-id))
         )
    (html:html
     (html:head
      (html:meta
       :http-equiv "content-type"
       :content #`"text/html; charset=,(cgi-output-character-encoding)")
      (html:meta :name "title"
                 :content (name-of title))
      (html:title (name-of title))
      )
     (html:body
      (html:h1 (name-of title))
      title-id
      ))))

(define (est-s-export-writer title-id subtitle-id)
  (let* ((root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         )
    (html:html
     (html:head
      (html:meta
       :http-equiv "content-type"
       :content #`"text/html; charset=,(cgi-output-character-encoding)")
      (html:meta :name "title"
                 :content (name-of title))
      (html:meta :name "subtitle"
                 :content (name-of subtitle))
      (html:title (name-of title) " / " (name-of subtitle))
      )
     (html:body
      (html:h1 (name-of title))
      title-id
      (html:p (name-of subtitle) " " subtitle-id)
      ))))

(define (est-a-export-writer title-id subtitle-id article-id)
  (let* ((root (get-root-data))
         (title (get-title root title-id))
         (subtitle (get-subtitle title subtitle-id))
         (article (get-article subtitle article-id))
         )
    (html:html
     (html:head
      (html:meta
       :http-equiv "content-type"
       :content #`"text/html; charset=,(cgi-output-character-encoding)")
      (html:meta :name "cdate"
                 :content (date->string (c-date-of article) date-format))
      (html:meta :name "mdate"
                 :content (date->string (m-date-of article) date-format))
      (html:meta :name "title"
                 :content (name-of title))
      (html:meta :name "subtitle"
                 :content (name-of subtitle))
      (html:title (if (string=? (header-of article) "")
                      "(no header)"
                      (regexp-replace #/^([^\n]*)(\n.*)?$/
                                      (header-of article)
                                      "\\1")))
      )
     (html:body
      (htmlize (header-of article))
      (htmlize (body-of article))
      (htmlize (footer-of article))
      (html:div (name-of title) " " title-id)
      (html:div (name-of subtitle) " " subtitle-id)
      ))))

(define (call-with-output-file-if-changed fname proc . args)
  (let ((tmpfname (if (page-tmp-path)
                      #`",(page-tmp-path)/,(sys-getpid).tmpout"
                      #`",|fname|.tmp"))
        (tmp2fname #`",|fname|.tmp2"))
    (dynamic-wind
        (lambda () #f)
        (lambda ()
          (apply call-with-output-file tmpfname
                 proc
                 args)
          (if (and (file-exists? fname)
                   (file-equal? tmpfname fname))
              (begin
                (sys-unlink tmpfname)
                #f)
              (begin
                (if (and (file-exists? fname)
                         (file-device=? tmpfname fname))
                    (sys-rename tmpfname fname)
                    (begin
                      (copy-file tmpfname tmp2fname :if-exists :supersede)
                      (sys-rename tmp2fname fname)
                      ))
                #t)))
        (lambda ()
          (sys-unlink tmpfname)
          (sys-unlink tmp2fname))
      )))


;; command => writer alist
(define cmd-alist
  `(("f" ,frame-writer #f)
    ("t" ,top-writer #f)
    ("i" ,index-writer #f)

    ("tl" ,tlist-writer #f)
    ("tls" ,tlist-sorted-writer #f)
    ("te" ,tedit-writer #f)
    ("tu" ,tupdate-writer #t)
    ("tm" ,tmove-writer #t)
    ("ta" ,tadd-writer #t)
    ("trc" ,trm-confirm-writer #f)
    ("tr" ,tremove-writer #t)
    ("tp" ,trelpage-writer #f)
    ("tpu" ,trelupdate-writer #t)
    ("tk" ,tkeywords-writer #t)

    ("sf" ,sframe-writer #f)
    ("sl" ,slist-writer #f)
    ("se" ,sedit-writer #f)
    ("su" ,supdate-writer #t)
    ("sm" ,smove-writer #t)
    ("sa" ,sadd-writer #t)
    ("src" ,srm-confirm-writer #f)
    ("sr" ,sremove-writer #t)
    ("sh" ,sheading-writer #t)
    ("shu" ,shupdate-writer #t)

    ("al" ,alist-writer #f)
    ("ae" ,aedit-writer #f)
    ("an" ,anew-writer #f)
    ("au" ,aupdate-writer #t)
    ("aa" ,aadd-writer #t)
    ("arc" ,arm-confirm-writer #f)
    ("ar" ,aremove-writer #t)

    ("r" ,rchange-writer #f)
    ("ra" ,rchange-atom-writer #f)

    ("ex" ,export-writer #f)
    ))


;; main

(define (aasummarizer-main . opt)
  (let-keywords* opt ((db-type <fsdbm>)
                      (db-path "aasdata")
                      (title "AA summarize")
                      (path "aasummarizer.cgi")
                      (base-url "")
                      (base-path "")
                      (tmp-path #f)
                      (export-path "aasexport")
                      (est-export-path #f)
                      (pre-export (lambda () '()))
                      (post-export (lambda () '()))
                      (css '())
                      (javascript '())
                      (default-style "aa")
                      (styles '())
                      (style-alist '())
                      (keywords #f)
                      (description #f)
                      (top-page default-top-page-writer)
                      (recent-days 20)
                      (recent-feed-days 7)
                      (pre-tlist (lambda args '()))
                      (post-tlist (lambda args '()))
                      (pre-alist (lambda args '()))
                      (post-alist (lambda args '()))
                      (encoding "Shift_JIS")
                      )
    (parameterize ((cgi-output-character-encoding encoding))
      (cgi-main
       (lambda (params)
         (parameterize ((cgi-params params)
                        (page-title title)
                        (page-path path)
                        (page-base-url base-url)
                        (page-base-path base-path)
                        (page-css css)
                        (page-js javascript)
                        (page-content-type "text/html")
                        (page-tmp-path tmp-path)
                        (aa-default-style default-style)
                        (aa-style-alist style-alist)
                        (page-keywords keywords)
                        (page-description description)
                        (top-page-writer top-page)
                        (export-dest export-path)
                        (est-export-dest est-export-path)
                        (pre-export-proc pre-export)
                        (post-export-proc post-export)
                        (rchange-days recent-days)
                        (rchange-feed-days recent-feed-days)
                        (pre-tlist-proc pre-tlist)
                        (post-tlist-proc post-tlist)
                        (pre-alist-proc pre-alist)
                        (post-alist-proc post-alist)
                        )
           (let* ((cmd (get-param "c" "f"))
                  (writer (assoc-ref cmd-alist cmd))
                  )
             (unless writer (error "Unknown command: " cmd))
             (with-db db-type db-path (cadr writer)
                      (lambda ()
                        (parameterize ((page-content-type (page-content-type)))
                          (let ((content ((car writer))))
                            `(,(cgi-header :content-type (page-content-type))
                              ,content)))))
             )))
       :on-error (lambda (e)
                   `(,(cgi-header)
                     ,(html:body
                       (html:pre
                        (html-escape-string
                         (call-with-output-string
                          (cut with-error-to-port <> (cut report-error e))))
                        ))))
       ))))


(provide "aasummarizer")
